#!/usr/bin/perl

package HTML::Phl;
our $VERSION = 0.00.001;  # 28.11.2017

use strict;
use warnings;
use utf8;  # Используем unicode
use CGI::Carp "fatalsToBrowser";  # Вывод ошибок в браузер
use Time::HiRes "gettimeofday";  # Для работы с microseconds
use Encode qw(from_to encode decode);  # Для работы с кодировками

our (%phl, %phl_var);

BEGIN {
  # Метка времени начала компиляции
  $phl_var{phl_t0} = gettimeofday;  # gettimeofday <= use Time::HiRes

  # Настройки по умолчанию (могут переопределяться в файле $phl{config})
  $phl{pl}     = 'phl';  # Расширение файлов PHL >> 'file_name.phl'
  $phl{config} = 'config.phl';  # Имя конфигурационного файла
  $phl{index}  = 'index.'.$phl{pl};  # Имя загружаемого файла по умолчанию

  $phl{top}    = '';  # Верхний колонтитул
  $phl{bottom} = '';  # Нижний колонтитул

  # Кодировка файлов: cp1251 utf8 koi8-r iso-8859-1 iso-8859-5 и др.
  $phl{encoding_in}  = "utf8";  # Кодировка содержимого загружаемых файлов
  $phl{encoding_out} = "utf8";  # Кодировка содержимого результирующих файлов
  $phl{locale}       = "utf8";  # Кодировка имен файлов и каталогов файловой системы

  $phl{header} = "Content-type: text/html\n\n";  # Заголовок документа

  $phl{lt} = '<%';  # начало блока инструкции
  $phl{rt} = '%>';  # конец блока инструкции
  $phl{ln} = '%';   # линейная инструкция
  $phl{sh} = 'pl';  # указатель принадлежности к языковой конструкции (в данном случае - к perl)

  $phl{config_path}  = '';     # Путь к конфигурационному файлу $phl{config}
  $phl{finde_config} = 'yes';  # Наличие поиска файлов конфигурации $phl{config}
  $phl{script_dir}   = '';     # Путь к директории возможного поиска загружаемых файлов *.cgi
  $phl{timer}        = 0;      # Таймер загрузки частей программы (детадизация в диаппазоне 0..4)
  $phl{bufer}        = 1;      # 1 >> '$| = 1';  0 >> '$| = 0'
  $phl{threads}      = '';     # Использование потоков (нитей) в рамках параллельного программирования

  # Загружаемые в процессе компиляции подпрограммы
  sub phl_timer {
     my $timer_msg   = $_[0];
     my $timer_file  = $_[1];

     &phl_encode($phl{encoding_out}, $timer_file);

     my $timer = '<p><small><font color=gray> • Время - </font>';
     
     if ($phl_var{phl_t1}) {
        $timer .= '<font color=#000099> '.phl_interval($phl_var{phl_t0}).'</font>'.
        '<font color=gray> ['.phl_interval($phl_var{phl_t1}).' ]</font> ';         
     } else {
        $timer .= '<font color=#000099> '.phl_interval($phl_var{phl_t0}).'</font>'.
        '<font color=gray> [ - ]</font> ';         
     }

     $timer .= '← <font color=#000099>'.$timer_msg.'</font> '.
               '<font color=gray>('.$timer_file.')</font></small></p>';

     if ($phl_var{header}  and  $phl_var{header} eq "yes"  and  !$_[3]) {
        $timer = encode("$phl{encoding_out}", $timer);
        print $timer."\n";
     } else {
        $timer = encode("$phl{encoding_out}", $timer);
        $phl_var{timer_eval} .= "print '$timer';\n";
     }
  }

  sub phl_interval {
     my $time = $_[0] || 0;
     return sprintf ("%.6f", (gettimeofday - $time));
  }

  sub phl_encode {
     my $encoding = shift;
     my $code = shift;

     if (!$encoding  and $phl{encoding_out} !~ /utf[-]{0,1}8/i) {
        $encoding = $phl{encoding_out};
     }

     if ($encoding  and  $encoding !~ /utf[-]{0,1}8/i) {
        if (ref($code)) {  # в подпрограмму передана ссылка
           $code = encode("$encoding", $$code);     
           return \$code;
        } else {
           $code = encode("$encoding", $code);     
        }
     }
        
     return $code;
  }
  
  sub phl_decode {
     my $decode = shift;
     my $content = shift;
     
     if ($decode eq "windows-1251") {
        $decode = "cp1251";
     }
     
     if ($decode !~ /utf[-]{0,1}8/i) {
        if (!$decode) {
           $content = decode("$phl{encoding_in}", $content);
        } elsif (ref($content)) {  # в подпрограмму передана ссылка
           $content = decode("$decode", $$content);
           return $$content;
        } else {
           $content = decode("$decode", $content);
        }
     } else {
        if (ref($content)) {  # в подпрограмму передана ссылка
           utf8::decode($$content);
           return $$content;
        } else {
           utf8::decode($content);
        }
     }
     return $content;   
  }

  # Выводим данные в зависимости от заданной $phl{encoding_out} кодировки
  sub phl_print ($)  {
     my $code = $_[0];
     
     my $encoding = $phl{encoding_out};
     
     if ($phl{encoding_out} eq "cp1251") {
        $encoding = "windows-1251";
     }
     
     print "Content-type: text/html; charset=$encoding\n\n"  if (!$phl_var{header});
     
     $phl_var{header} = "yes";
     print phl_encode("", $code);
  }

  sub phl_fileparse {
     my $path_to_file = $_[0];

     my ($dirpath, $file_name, $separator, $suffice);

     $path_to_file =~ /^(.*?[^\?]*(\/|\\)){0,1}(.*?[^\?]*)(\?{0,1}.*?)\s*$/;

     $dirpath = $1 || '';
     $separator = $2;
     $file_name = $3;
     $suffice = $4;
     
     if ($dirpath eq "http://") {
        $dirpath .= $file_name."/";
        $file_name = "";
     }

     if ($dirpath =~ /[\?*"<>|]/) {
        &phl_error("Ошибка загрузки \$dirpath = '$dirpath' : '$path_to_file'");
     } elsif ($file_name =~ /[\?*"<>|:]/) {
        &phl_error("Ошибка загрузки \$file_name = '$file_name' : '$path_to_file'");
     }

     if (wantarray()) {  # Списковый контекст
        return ($dirpath, $file_name, $separator, $suffice);
     } else {  # Неопределенный или скалярный контекст
        return $file_name;
     }

  }
  
  # Загружаем основной конфигурационный файл "$phl{config}", переопределяющий настройки, принятые по умолчанию
  &phl_timer("Начало загрузки основного файла config", "$phl{config}")  if ($phl{timer} >= 4);

  require "$phl{config_path}$phl{config}"  if (-e "$phl{config_path}$phl{config}");

  &phl_timer("Конец загрузки основного файла config", "$phl{config}")  if ($phl{timer} >= 4);
}
BEGIN { 

  # Поиск дополнительных файлов config
  my ($file_path, $file_name);
   if ($ENV{PATH_TRANSLATED} =~ /\?/) {
     ($file_path, $file_name) = &phl_fileparse($ENV{DOCUMENT_ROOT}.$ENV{REQUEST_URI});
     $file_path =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
     $file_name =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
  } else {
     ($file_path, $file_name) = &phl_fileparse($ENV{PATH_TRANSLATED});
  }

  # Ищем конфигурационный файл "$phl{config}" в каталоге, содержащем $ENV{PHL_PATH_TRANSLATED}
  my $config_path = $file_path;
  while ($phl{finde_config}) {
     if (-e $config_path.$phl{config}) {
        # Загружаем конфигурационный файл "$phl{config}"
        require $config_path.$phl{config};
        last;
     } elsif ($config_path =~ s/(.*)(\/|\\)(?:\w)+(?:\/|\\)$/$1$2/gi) {
        (my $local_config_path = $config_path) =~ s/(?:\/|\\)//gi;
        (my $local_DOCUMENT_ROOT = $ENV{DOCUMENT_ROOT}) =~ s/(?:\/|\\)//gi;
        last if ($local_config_path eq $local_DOCUMENT_ROOT);  # Дальше не ищем файл $phl{config}
     } else {
        last;
     }
  }
  &phl_timer("Конец поиска дополнительных файлов config", "$phl{config}")  if ($phl{timer} >= 4);
}

# Модули, используемые в работе PHL (распределены по программе, загружаясь при необходимости) - под управлением mod_perl лучше загружать при компиляции
use threads;  # Работа с потоками (если поддерживает сборка perl)
# use IO::Socket::INET;
# use IO::Socket::SSL;
# use URI;

BEGIN { 
  &phl_timer("Конец загрузки части стандартных модулей", "use Modules perl")  if ($phl{timer} >= 3);

  (my $new_INC = $ENV{SCRIPT_FILENAME}) =~ s|/HTML/.*||;
  unshift (@INC, "$new_INC");

  &phl_timer("Корректировка \@INC", "$new_INC")  if ($phl{timer} >= 4);
}

# Модули-расширения PHL. Можно загружать в 'config.phl'
use HTML::Phl::Psimple;
use HTML::Phl::Utilit;
use HTML::Phl::Php;
use HTML::Phl::Ru;
use HTML::Phl::Http;

BEGIN {
  &phl_timer("Конец загрузки модулей PHL", "end Modules")  if ($phl{timer} >= 3);
}

# CHECK { # CHECK может не сработать с mod_perl, так что засечка времени завершения компиляции сделана в конце программы в блоке BEGINE {}
  # Метка времени конца компиляции программы
  # $phl_var{phl_t1} = gettimeofday;  # gettimeofday <= use Time::HiRes
  # &phl_timer("CHECK - конец компиляции", "Phl.pm")  if ($phl{timer} >= 2);
# }

END {
  &phl_timer("PROGRAM_END", "$ENV{PHL_PATH_TRANSLATED}) ($phl_var{name})")  if ($phl{timer} >= 1);
}

# ############################ BEGIN_PROGRAM ################################# #

&phl_timer("BEGIN_PROGRAM", "$ENV{PHL_PATH_TRANSLATED}")  if ($phl{timer} >= 2);

$| = 1  if ($phl{bufer});  # Сброс буфера

my (%phl_threads, %phl_fork, %phl_join, @phl_blok);

# Определяем имя файла и путь к директории исполняемого файла *.$phl{pl} для дальнейшего его анализа, в части поиска частей кода perl
my ($file_name_uri, $separator, $suffice_uri) = (&phl_fileparse($ENV{REQUEST_URI}))[1, 2, 3];
$suffice_uri =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
$file_name_uri =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
from_to($file_name_uri, "utf8", "$phl{locale}");
$ENV{PHL_REQUEST_URI} = $separator.$file_name_uri.$suffice_uri;

# $ENV{PATH_TRANSLATED} >> $ENV{PHL_PATH_TRANSLATED}
my ($file_path, $file_name);
if ($ENV{PATH_TRANSLATED} =~ /\?/) {
   ($file_path, $file_name) = &phl_fileparse($ENV{DOCUMENT_ROOT}.$ENV{REQUEST_URI});
   $file_path =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
   $file_name =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
} else {
   ($file_path, $file_name) = &phl_fileparse($ENV{PATH_TRANSLATED});
}

my $file_url = "http://$ENV{SERVER_NAME}$ENV{PATH_INFO}";
my $url_dir  = (phl_fileparse("$file_url"))[0];

$ENV{PHL_PATH_TRANSLATED} = $file_path.$file_name;

if ($phl{locale} and $phl{locale} ne "utf8") {
   # Перекодируем запрос в $phl{locale}, т.к. имена файлов и каталогов кодируются браузером в кодировке UTF-8
   from_to($ENV{PHL_PATH_TRANSLATED}, "utf8", "$phl{locale}");
} else {
   # Поднимаем флаг, если имена файловой системы в кодировке UTF-8
   utf8::decode($ENV{PHL_PATH_TRANSLATED});
}

($file_path, $file_name) = &phl_fileparse("$ENV{PHL_PATH_TRANSLATED}");

$file_name = $phl{index}  if (!$file_name);

# Завершаем выполнение программы, если указан конфигурационный файл $phl{config}
exit  if ($file_name eq $phl{config});

# Запоминаем переменную окружения
my @phl_query_string;
$phl_var{'QUERY_STRING'} = $ENV{'QUERY_STRING'};
push @phl_query_string, $phl_var{'QUERY_STRING'};

# Изменяем текущий каталог процесса
chdir "$file_path" or die "Невозможно cd в $file_path";

# Запоминаем данные по текущему каталогу
our @phl_path_process;
push @phl_path_process, $file_path;

# Запоминаем данные по текущему url
our @phl_path_url;
push @phl_path_url, $url_dir;

# Запоминаем данные по текущей кодировке файлов
our @phl_files_decode;
push @phl_files_decode, "$phl{encoding_in}";

# Формируем хэш экспорта ключей и параметров из сторонних модулей
our %phl_export;
&phl_export();

my ($phl_analys_top, $phl_analys_main, $phl_analys_bottom) = ('', '', '');

# Разбираем структуру верхнего колонтитула (если такой задан в файле "$phl{config}")
$phl_analys_top = &phl_analys_code("$phl{config} (top)", \$phl{top})  if ($phl{top});

# Разбираем структуру файла *.$phl{pl} на предмет поиска perl-кода и text/html
$phl_analys_main = &phl_analys_file($file_path.$file_name, "", "$file_url");

# Разбираем структуру нижнего колонтитула (если такой задан в файле "$phl{config}")
$phl_analys_bottom = &phl_analys_code("$phl{config} (bottom)", \$phl{bottom})  if ($phl{bottom});

# Анализируем данные, посланные методом GET или POST
my %phl_query;
&phl_query("main");

# Формируем заголовок, если не запрещено в конфигурационном файле
&phl_header();

eval "$phl_var{timer_eval}";
&phl_code_perl_eval("$phl_analys_top"."$phl_analys_main"."$phl_analys_bottom", "$phl_var{phl}", "", "$file_name");

# Формируем заголовок
sub phl_header {

   if ($phl_var{header} and !$_[0]) { # заголовок не нужен "no" или уже есть "yes"
   
   } elsif ($phl{header}) {
      # Если не задана в заголовке кодировка, осуществляется попытка это исправить
      if ($phl{header} =~ m|text/html|i  and $phl{header} !~ /charset/i) {
         $phl{encoding_out} = "utf8"  if (!$phl{encoding_out});
         $phl{header} = "Content-type: text/html; charset=$phl{encoding_out}\n\n";
      }
      print $phl{header};
      $phl_var{header} = "yes"; # метка наличия заголовка
   }
}

# Подпрограмма формирования хэша экспорта ключей и параметров из сторонних модулей
sub phl_export {
   foreach my $key (keys (%main::HTML::Phl::)) {
   
      no strict "refs";

      foreach my $m_key (keys (%{"main::HTML::Phl::$key"})) {
         if ($m_key eq "phl_import") {
            foreach my $modul_key (keys (%{"HTML::Phl::"."$key"."phl_import"})) {
               foreach my $code_key (keys (%{${"HTML::Phl::"."$key"."phl_import"}{$modul_key}})) {
                  # Формируем хэш экспорта из доп.модулей
                  $phl_export{$modul_key}{$code_key} = ${"HTML::Phl::"."$key"."phl_import"}{$modul_key}{$code_key};
               }
            }
         }
      }
   }

}

# Анализатор структуры файла HTML + code Perl
sub phl_analys_file {  # $_[0] - разбираемый файл с частями кода Perl,  $_[1] - параметры

   my $analys_file = $_[0];
   my $param = $_[1];
   my $file_url = $_[2];

   &phl_timer("OPEN_FILE", "$analys_file")  if ($phl{timer} >= 4);

   my @lines;
   open(my $fh, "<", "$analys_file") or &phl_error($analys_file, '$!');
      flock($fh, 1);
      my ($bate, $code);
      while (sysread($fh, $bate, 8*1024)) {
         $code .= $bate;
      }
   close($fh);

   &phl_timer("CLOSE_FILE", "$analys_file")  if ($phl{timer} >= 4);

   # Перекодируем входные данные в зависимости от заданной $phl{encoding_in} кодировки
   my $decode;
   if ($param =~ /(?:dc|decode)(?:=([\w\-]*)){0,1}/) {
      $decode = $1;
      if (!$decode) {
         $decode = $phl{encoding_in};
      }
   } else {
      $decode = $phl{encoding_in};
   }

   if ($decode eq "utf8") {
       # Если в файле находиться данные в UTF-8 - просто поднимаем флаг
       utf8::decode($code);
   } else {
       $code = decode($decode, $code);
   }

   &phl_timer("END_DECODE_FILE", "$analys_file")  if ($phl{timer} >= 4);

   return &phl_analys_code($analys_file, \$code, "$file_url");  # Отправляем данные из файла в анализатор
}

# Анализатор
sub phl_analys_code {

   my $analys_file = $_[0];
   my $lines_ref = $_[1];
   my $file_url = $_[2];

   &phl_timer("ANALIZATOR_CODE_BEGINE", "$analys_file")  if ($phl{timer} >= 4);

   my ($code_perl, $code_html, $finde_code_perl, $finde_code_perl_all) = ('', '', '', '');
   my $nota_blok = '';
   my ($sh_blok, $key_blok, $name_blok, $code_perl_blok);
   my $number_blok = -1;
   foreach my $line (split(/\r?\n/m, $$lines_ref)) {

      utf8::decode($line)  if ($phl{encoding_in} eq "utf8");

      # Инструкция '<%pl code_perl %>'
      if ($line  and  $finde_code_perl ne "yes"  and  $line =~ /\Q$phl{lt}\E/  and  $line =~ /^(.*?)\Q$phl{lt}\E([^ ]*?)(?:(=|:)([^ ]*)){0,1}(?: +(\-[^ ]*)){0,1}(?: +(.*?) +){0,1} *\Q$phl{rt}\E(.*)$/) {
         my $code_html_begine = $1;
         my $sh = $2;
         my $nota = $3 || '';
         my $name_blok = $4;  # Не используется
         my $key = $5;
         my $code = "$6";
         my $code_html_end = $7;   # ???
            $number_blok++;
            $phl_blok[$number_blok] = "$name_blok";

         $code_perl .= &phl_esc($code_html)  if ($code_html);
         $code_perl .= &phl_esc($code_html_begine."\n", "print")  if ($code_html_begine);

         # Если $sh отличен от указателя $phl{sh} на инструкцию, содержащую код perl, - видоизменяем $code
         if ($sh  and  $sh ne $phl{sh}  and  &phl_finde_sh($sh)) {
            $code = &phl_share($code, $sh, $nota, $key, "$analys_file", "$file_url");
            # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
            $code_perl .= ($key) ? &phl_key($key, $code."\n", $number_blok) : $code."\n";  
         } elsif ($sh eq $phl{sh}) {
            if ($nota eq '=') {
               $code =~ s/;+$//;
               $code = "print($code);"  
            } else {
               $code .= ";"  if ($code !~ /;\s*$/);            
            }
            # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
            $code_perl .= ($key) ? &phl_key($key, $code."\n", $number_blok) : $code."\n";  
         } else {
            $code_perl .= &phl_esc("$phl{lt}$sh$nota$name_blok $key $code $phl{rt}", "print");
         }

         if ($code_html_end) {
            # Находим последующие конструкции  <%pl code ... %>  в данной строке
            while ($code_html_end =~ m/(.*?)\Q$phl{lt}\E([^ ]*?)(?:(=|:)([^ ]*)){0,1}(?: +(\-[^ ]*)){0,1}(?: +(.*?) +){0,1} *\Q$phl{rt}\E/gc) {
               my $code_html_begine = $1;
               my $sh = $2;
               my $nota = $3;
               my $name_blok = $4;
               my $key = $5;
               my $code = $6;
               $number_blok++;
               $phl_blok[$number_blok] = "$name_blok";

               $code_perl .= &phl_esc($code_html_begine."\n", "print");

               # Если $sh отличен от указателя $phl{sh} на инструкцию, содержащую код perl, - видоизменяем $code
               if ($sh  and  $sh ne $phl{sh}  and  &phl_finde_sh($sh)) {
                  $code = &phl_share($code, $sh, $nota, $key, "$analys_file", "$file_url");
                  $code_perl .= ($key) ? &phl_key($key, $code."\n") : $code."\n";  # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
               } elsif ($sh eq $phl{sh}) {
                  if ($nota eq '=') {
                     $code =~ s/;+$//;
                     $code = "print($code);"  
                  } else {
                     $code .= ";"  if ($code !~ /;\s*$/);            
                  }
                  $code_perl .= ($key) ? &phl_key($key, $code."\n") : $code."\n";  # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
               } else {
                  $code_perl .= &phl_esc("$phl{lt}$sh$nota$name_blok $key $code $phl{rt}", "print");
               }

            }

            if ($code_html_end =~ /(.*)$/g) {
               $code_html_end = $1;
               $code_perl .= &phl_esc($code_html_end."\n");
            }

         } else {
            $code_perl .= "\n";
         }

         $code_html = '';  #  Присваиваем новое (последнее) значение

      }  # Начало инструкции '<%pl code ...'
      elsif ($line  and  $finde_code_perl ne "yes"  and  $line =~ /\Q$phl{lt}\E/  and  $line =~ /^(.*)\Q$phl{lt}\E([^ \n:=]*)(=|:){0,1}([^ \n]*) *(\-[^ \n]+)* *(.*)$/) {

         my $code_html_begine = $1;
            $sh_blok = $2;
            $nota_blok = $3;
            $name_blok = $4;
            $key_blok = $5;
         my $code = $6;
         
         $number_blok++;
         $phl_blok[$number_blok] = "$name_blok";

         $code_perl .= &phl_esc($code_html);
         $code_html  = ""; #
         $code_perl .= &phl_esc($code_html_begine."\n", "print")  if ($code_html_begine  and  $code_html_begine !~ /^[ ]+$/);
         $code_perl_blok .= $code."\n"  if ($code);
         $finde_code_perl = "yes";

      } # Конец инструкции '... code %>'
      elsif ($line  and  $finde_code_perl eq "yes"  and  $line =~ /\Q$phl{rt}\E/  and  $line =~ /^(.*?)\Q$phl{rt}\E(.*)$/) {

         my $code = $1;
         my $code_html_end = $2;

         $code_perl_blok .= $code  if ($code  and  $code !~ /^[ ]+$/);  # code perl
        
         # Если $sh отличен от указателя $phl{sh} на инструкцию, содержащую код perl, - видоизменяем $code
                           
         if ($sh_blok  and  $sh_blok ne $phl{sh}  and  &phl_finde_sh($sh_blok)) {
            $code_perl_blok = &phl_share($code_perl_blok, $sh_blok, $nota_blok, $key_blok, "$analys_file", "$file_url");  
            # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
            $code_perl .= &phl_key($key_blok, $code_perl_blok, $number_blok, $name_blok);  
            $code_perl .= "\n"  if ($code  and  $code !~ /^[ ]+$/);            
         } elsif ($sh_blok eq $phl{sh}) {
            if ($nota_blok  and  $nota_blok eq '=') {
               $code_perl_blok = &phl_esc($code_perl_blok, '=') 
            }
            # Если указан ключ $key то видоизменяем код $code в подпрограмме &phl_key
            $code_perl .= &phl_key($key_blok, $code_perl_blok, $number_blok, $name_blok);  
         } else {
            $code_perl .= &phl_esc("$phl{lt}$sh_blok$nota_blok$name_blok $key_blok $code_perl_blok$phl{rt}", "");
         }
          
         $code_perl .= &phl_esc($code_html_end."\n", "")  if ($code_html_end  and  $code_html_end !~ /^[ ]+$/);
         $sh_blok   = '';
         $nota_blok = '';
         $name_blok = '';
         $key_blok  = '';
         $code_perl_blok = '';
         $code_html = '';
         $finde_code_perl = "no";

      }  # Инструкция '%pl code_perl ...'
      elsif ($line  and  $finde_code_perl ne "yes"  and  $line =~ /^\Q$phl{ln}\E/  and  $line =~ /^\Q$phl{ln}\E([^ \n]*)(=|:){0,1}([^ \n]*) *(\-[^ \n]+)* *(.*)$/) {

         my $sh        = $1;  # расширение инструкции (принято 'pl' - для perl)
         my $nota      = $2;  # не используется в данной конструкции
         my $name_blok = $3;  # не используется в данной конструкции
         my $key_blok  = $4;
         my $code      = $5;
         
         $code_perl .= ($code_html) ? &phl_esc($code_html) : "";
         $code_html = "";
               
         # Если $sh отличен от указателя $phl{sh} на инструкцию, содержащую код perl, - видоизменяем $code
         if ( ($sh eq $phl{sh})  or  &phl_finde_sh($sh) ) {
            if ($sh eq $phl{sh}) {
               $code .= ";"  if ($code !~ /;\s*$/);  
            } else {
               $code = &phl_share($code, $sh, $nota, $key_blok, "$analys_file", "$file_url");
            }
            
            if ($key_blok  and  $key_blok =~ /-(?:pl|perl_all)/) {
               $finde_code_perl_all = "yes";
               $code = '';
               $phl_var{header} = 'no';
               $phl_var{timer_eval} = '';
               $phl{bottom} = '';
               $phl{top} = '';
            } else {
               $code .= "\n"  if ($code);
               # Если указан ключ $key то видоизменяем $code в подпрограмме &phl_key
               $code_perl .= ($key_blok) ? &phl_key("$key_blok", "$code", "line") : $code;
            }

         } else {
            $code_perl .= &phl_esc("$phl{ln}$sh$nota$name_blok $key_blok $code", "");
         }
         
      } else {

         if ($finde_code_perl eq "yes") {  # ... code ...
            $code_perl_blok .= $line."\n";  # code perl
         } elsif ($finde_code_perl_all eq "yes") {  # %pl code_perl ...
            $code_perl .= $line."\n";  # code perl
         } else {  # ... html ...
            $code_html .= $line."\n";
         }

      }

   }

   if ($finde_code_perl_all eq "yes") {
      $code_perl = &phl_key($key_blok, $code_perl, 1, $name_blok);
   }

   $code_perl .= &phl_esc($code_html)  if ($code_html  and  $code_html ne "\n");

   &phl_timer("ANALIZATOR_CODE_END", "$analys_file")  if ($phl{timer} >= 4);

   return $code_perl;
}

sub phl_finde_sh {
   my $key_sh = shift;
   if ($phl_export{sh}{$key_sh}) {  # если указатель инстукции совпадает с ключем хэша %phl_export
      return "finde_sh";
   } else {
      return "";
   }
}

# Расширение инструкций (указатели, отличные от perl)
sub phl_share {
   my $code = $_[0];
   my $sh   = $_[1];
   my $nota = $_[2];
   my $key  = $_[3];
   my $analys_file = $_[4]; 

   $code = $phl_export{sh}{$sh} -> ("$code", "$sh", "$nota", "$key", "$analys_file");  # передаем управление подпрограмме, указанной в %phl_export
   
   return $code;
}

# Управляющие ключи
sub phl_key {
  my $key = $_[0];
  my $code = $_[1];
  my $number_blok = $_[2];  # Номер блока или "line"
  my $name_blok   = $_[3];  # Если указано имя блока <%pl:name_blok
  
  my ($name_process, $name_var);
  
  if ($key =~ /-(?:ev|eval)\b/) {
     eval "$code";
     
     # Вывод ошибок
     &phl_error($number_blok, '$@')  if ($@);
     
     &phl_export();  # Обновляем хэш экспорта ключей из сторонних модулей
     return;
  }

  # Остановка или запуск таймера

  if ($key =~ /-(?:fk|fork)\b(?:=([^ \n,\-]*)){0,1}/) {
     $name_process = $1;
     ($name_process, $name_var) = &phl_key_process($name_process, $name_blok);
     $phl_var{$name_var} = $code;
     $code = "  \$phl_join{$name_process} = &phl_fork(\$phl_var{$name_var}, 'code', '$name_process');\n";

  } elsif ($key =~ /-(?:td|thread)\b(?:=([^ \n,\-]*)){0,1}/) {
     $name_process = $1;
     ($name_process, $name_var) = &phl_key_process($name_process, $name_blok);
     $phl_var{$name_var} = $code;
     $code = "  \$phl_join{$name_process} = &phl_thread(\$phl_var{$name_var}, 'code', '$name_process');\n";

  } elsif ($key =~ /-(?:jn|join)\b(?:=([^ \n\-]*)){0,1}/) {  # <%pl -join=t1 ...   # <%pl -join=t1,t2,t5 ...   # <%pl -join ...
     $name_process = $1;
     $name_process =~ s/,+/,/g;
     $name_process =~ s/,+$//;
     $name_process =~ s/^,//;
     my $result_hash;
     ($result_hash = $name_process) =~ s/([^,]+)/\$phl_join{$1}/g;  # >> $phl_join{$name_process}
     if (!$name_process) {
        $result_hash = "%phl_join";
     } else {
        $result_hash = "($result_hash)";
     }
     $code = "  $result_hash = &phl_end_process('$name_process', 'code');\n$code\n";
  }

  if ($key =~ /-(?:bg|begin)\b/) {
     $code = "BEGIN {\n $code \n}\n";
  } elsif ($key =~ /-end\b/) {
     $code = "END {\n $code \n}\n";
  } elsif ($key =~ /-(?:ch|check)\b/) {
     $code = "CHECK {\n $code \n}\n";
  }

  # Поиск ключей от сторонних модулей

  foreach my $finde_key (split('-', $key)) {
     if ($phl_export{key}{$finde_key}) {
        # передаем управление подпрограмме, указанной в %phl_export
        $code = $phl_export{key}{$finde_key} -> ("$code", "$number_blok", "$key");
     }
  }
            
  return $code;
}

sub phl_key_process {
  my $name_process = $_[0];
  my $name_blok = $_[1];
  my $name_var;

  if ($name_blok) {
     if ($name_process) {  # <%pl:name_blok -td=name ...   # <%pl:join -fk=name ...
        $name_var = ($name_blok eq "join") ? "$name_process" : "$name_blok";
     } else {  # <%pl:name_blok -td ...  # <%pl:join -td ...
        $name_var = ($name_blok eq "join") ? gettimeofday : "$name_blok";  # gettimeofday <= use Time::HiRes
        # Может при <%pl:join -td ... закрывать все открытые нити?
     }
     $name_process = $name_blok;
  } else {
     if ($name_process) {  # <%pl -fk=name ...  # <%pl -td=join ...
        $name_var = ($name_process eq "join") ? gettimeofday : "$name_process";  # gettimeofday <= use Time::HiRes
     } else {  # <%pl -td ...
        $name_process = "join";
        $name_var = gettimeofday;  # gettimeofday <= use Time::HiRes
     }
  }
  return ($name_process, $name_var);
}

# HTML в Perl
sub phl_esc {
  my $esc = $_[0];
  my $nota = $_[1] || '';
  
  $esc .= "\n"  if ($esc !~ /\n$/);

  if ($nota eq '=') {
     return " print <<\"[HTML]\";\n${esc}[HTML]\n";
  } elsif ($nota eq 'print') {
     return " print <<'[HTML]' =~ /(.*)/;\n${esc}[HTML]\n";  # Для исключения пробелов между print'ами
  } else {
     return " print <<'[HTML]';\n${esc}[HTML]\n";
  }
}

# Выполняем найденный в HTML код Perl
sub phl_code_perl_eval {

  my $code = shift;  # код
  my $param = shift;
  
  # Для доступа в файле include к имени процесса
  $phl_var{name} = shift;
  $phl_var{name} = "main"  if (!$phl_var{name});
  
  my $file_include = shift;
  
  &phl_timer("EVAL_BEGINE", "$file_include")  if ($phl{timer} >= 4);

  $code = "no strict;\n".$code."use strict;\n"  if ($param =~ /\b(?:ns|no_strict)\b/);

  $param .= " $phl_var{phl}";  # для возможности добавления в параметры управляющих элементов из модулей
  
  # Поиск параметров от сторонних модулей
  my $finde_param;
  foreach my $name_param (split(' ', $param)) {
     if ($phl_export{eval}{$name_param}) {
        $finde_param = $name_param;
        last;
     }
  }
  
  my @phl_result;

  if ($finde_param  and  $phl_export{eval}{$finde_param}) {
     # передаем управление подпрограмме, указанной в %phl_export
     $code = $phl_export{eval}{$finde_param} -> ("$code", "$param", "$phl_var{name}", "$file_include");
     return if (!$code); # Если нет возврата измененного кода не отправляем на eval

  } elsif ($param =~ /\b(?:ne|no_eval)\b/) {
     # ...
     # ничего не делаем - код не выполненяется
     # ...
  } else {
     # Перекодируем выходные данные в зависимости от $phl{encoding_out}

     if ($phl{encoding_out} eq "utf8") {
        utf8::decode($code);
     } else {
        $code = encode($phl{encoding_out}, $code);
        utf8::encode($code);
     }

     if (wantarray()) {  # Списковый контекст
        @phl_result = (eval $code);
     } elsif (defined wantarray()) {  # Скалярный контекст
        @phl_result = (eval $code); # ???
     } else {  # Неопределенный контекст
        eval $code;
     }

  }

  # Вывод ошибок
  &phl_error($file_include, '$@')  if ($@);

  &phl_timer("EVAL_END", "$file_include")  if ($phl{timer} >= 3);

  return @phl_result;
}

sub phl_error {
  my $error_msg = $_[0];
  my $error_label = $_[1];

  &phl_header();

  print "<hr size=1 width=100 align=left color=blue><font color=red><i>";

  my ($package, $filename, $line, $subroutine, $hasargs, $wantargs) = caller();

  if ($error_label eq '$@') {
     $@ =~ s/\n/<br>\n/g;
     phl_print "<B>eval error:</B> '$error_msg'<BR>$@ : $package, $filename, $line, $subroutine, $hasargs, $wantargs";
  } elsif ($error_label eq '$!') {
     phl_print "'$error_msg': $! : $package, $filename, $line, $subroutine, $hasargs, $wantargs";
  } else {
     phl_print "'$error_msg': $package, $filename, $line, $subroutine, $hasargs, $wantargs";
  }

  print "</i></font><hr size=1 width=100 align=left color=blue>";
}

# Разбираем переданные файлу данные (POST и GET), заполняя хеш '%phl_query'
sub phl_query {

   my ($qs, $decode);
   if ($ENV{'REQUEST_METHOD'} eq "GET") {
      $qs = $phl_var{'QUERY_STRING'};
   } elsif ($ENV{'REQUEST_METHOD'} eq "POST") {
      read(STDIN, $qs, $ENV{'CONTENT_LENGTH'});
   }

   # для файлов, подключаемых по 'include', для формирования %phl_query используем данные QUERY_STRING, если данные по POST отсутствуют
   $qs = $phl_var{'QUERY_STRING'}  if (!$qs and $phl_var{'QUERY_STRING'});  
   
   my @split = split (/&/, $qs);
   foreach my $split (@split) {
      my ($query_key, $query_value) = split (/=/, $split);
      $query_value =~ tr/+/ /;
      $query_value =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;
      $query_key =~ tr/+/ /;
      $query_key =~ s/\?//;  # удаляем знак "?"
      $query_key =~ s/%([\dA-Fa-f]{2})/pack("C", hex($1))/eg;

      # Декодируем данные POST/GET
      # Переводим входные данные UTF-8 во внутренний формат perl (с флагом)
      utf8::decode($query_key);
      utf8::decode($query_value);      

      $phl_query{$query_key} = $query_value;

   }
   return;
}

# fork in include
sub phl_fork {

   my $code_or_name_file = $_[0];  # $_[0] - включаемый в тело документа код или файл
   my $param     = $_[1];
   my $name_fork = $_[2];

   my $print_code_or_name_file = ($param =~ /\bcode\b/) ? "code (name=$name_fork)" : "$code_or_name_file (name=$name_fork)";

   &phl_timer("FORK_BEGINE", "$print_code_or_name_file")  if ($phl{timer} >= 2);

   $name_fork = "waitpid"  if (!$name_fork  or  $name_fork eq "join");
   $name_fork = "no_waitpid"  if ($param =~ /\b(?:nw|no_waitpid)\b/);

   my ($fh_READER, $fh_WRITER);
   if ($name_fork ne "no_waitpid") {
      pipe $fh_READER, $fh_WRITER;  # Создаем программный канал

      # Запоминаем в хеше и привязываем к имени дочернего процесса соответствующие дескрипторы канала чтения данных из него
      $phl_fork{$name_fork} = $fh_READER;  
   }

   # 'Разветвить'  текущий процесс, fork вернет '0' в порожденный процесс (потомок)
   my $pid = fork();
   die "fork не сработал: $!"  unless (defined $pid);  # обработка ошибок

   # порожденный процесс (потомок)
   if ($pid == 0) {
      close $fh_READER  if ($name_fork ne "no_waitpid");

      my @result;
      if ($param =~ /\bcode\b/) {
         @result = &phl_code_perl_eval($code_or_name_file, $param, $name_fork);
      } else {
         @result = &include($code_or_name_file, $param, $name_fork);  # Возвращаемся обратно в &include, но уже в порожденном потомке
      }

      # Пишем в программный канал
      if ($name_fork ne "no_waitpid") {
         foreach my $result (@result) {

            print $fh_WRITER $result."\n";
         }
         close $fh_WRITER;
         exit;
      } else {
         exit;
      }
   }

   close $fh_WRITER  if ($name_fork ne "no_waitpid");

   # Ждём завершения потомка-процесса, если только указано "waitpid" или отсутствует имя порожденного процесса
   if ($name_fork eq "waitpid") {
      waitpid($pid, 0);

      # Считываем из программного канала
      my @result = <$fh_READER>;
      chomp(@result);

      close $fh_READER  if (!$name_fork);

      delete $phl_fork{$name_fork};

      &phl_timer("FORK_END", "$print_code_or_name_file")  if ($phl{timer} >= 2);

      # Переменной ставим флаг юникода, если данные utf-8
      if ($phl{encoding_out} eq "utf8") {
         utf8::decode($_)  foreach (@result);
      }

      if (wantarray()) {  # Списковый контекст
         return @result;
      } else {  # Неопределенный или Скалярный контекст
         return ($#result) ? \@result : $result[0];
      }

   }

}

# thread blok & include
sub phl_thread {

   eval 'use threads'  if ($phl{threads});  # use threads;

   my $code_or_name_file = $_[0];  # $_[0] - включаемый в тело документа код (в виде хеша) или файл
   my $param             = $_[1];
   my $name_thread       = $_[2];

   my $print_code_or_name_file = ($param =~ /\bcode\b/) ? "code (name=$name_thread)" : "$code_or_name_file (name=$name_thread)";

   &phl_timer("THREAD_BEGINE", "$print_code_or_name_file")  if ($phl{timer} >= 2);

   $name_thread = "join"    if (!$name_thread  or  $param =~ /\bjoin\b/);
   $name_thread = "detach"  if ($param =~ /\b(?:dt|detach)\b/);

   ($phl_threads{$name_thread}) = threads->create(
      sub {  # create() <= use threads
         my @result;
         if ($param =~ /\bcode\b/) {
            @result = &phl_code_perl_eval("$code_or_name_file", "$param", "$name_thread");
         } else {
            # Возвращаемся обратно в &include, но уже в порожденном потомке
            @result = &include($code_or_name_file, $param, $name_thread);
         }
         # phl_print "1: \@result=@result <br>", join(" : ", @result), "<br>";
         return @result;
      }
   );

   if ($name_thread eq "join"  or  $name_thread eq "detach") {
      my @result = ();
      @result = $phl_threads{$name_thread}->join()  if ($name_thread eq "join");    #  join() <= use threads
      $phl_threads{$name_thread}->detach()          if ($name_thread eq "detach");  #  detach() <= use threads

      delete $phl_threads{$name_thread};

      &phl_timer("THREAD_END", "$print_code_or_name_file")  if ($phl{timer} >= 2);

      if ($name_thread ne "detach") {
         if (wantarray()) {  # Списковый контекст
            return @result;
         } else {  # Неопределенный или Скалярный контекст
            return ($#result) ? \@result : $result[0];
         }
      }

   }

}

sub phl_end_process {

  my $name_process = $_[0];
  my $wantarray = $_[1];

     my %result;
     if ($name_process) {
        my @result = ();
        my @threads_or_fork = split(' ', $name_process);
        if (@threads_or_fork) {  # Несколько процессов (нитей или fork) при их указании
           foreach my $name (@threads_or_fork) {
              # Формируем массив (и хеш) результатов выполнения выбраных ($name) нитей или fork
              if ($phl_threads{$name}) {  # Найдена нить

                 my @result_fh = $phl_threads{$name}->join();  # join() <= use threads
                 $result{$name} = ($#result_fh) ? \@result_fh : $result_fh[0];

                 push @result, $result{$name};

                 delete $phl_threads{$name};
                 &phl_timer("THREAD_END", "$name")  if ($phl{timer} >= 2);
              } elsif ($phl_fork{$name}) {  # Найден дочерний процесс

                 # Считываем из программного канала

                 my $fh = $phl_fork{$name};  # Считываем дескриптор канала чтения данных в зависимости от имени дочернего процеса
                 my @result_fh = <$fh>;
                 close $fh;

                 # Переменной ставим флаг юникода, если данные utf-8
                 if ($phl{encoding_out} eq "utf8") {
                    utf8::decode($_)  foreach (@result_fh);
                 }

                 $result{$name} = ($#result_fh) ? \@result_fh : $result_fh[0];

                 push @result, $result{$name};

                 delete $phl_fork{$name};
                 &phl_timer("FORK_END", "$name")  if ($phl{timer} >= 2);
              } else {
                 &phl_error("Процесс <B>'$name'</B> не найден<br>");
              }
           }
        } else {  # Одна нить или дочерний процесс
           my $name = $name_process;
           if ($phl_threads{$name}) {  # Найдена нить

              @result = $phl_threads{$name}->join();  # join() <= use threads
              $result{$name} = "@result";

              delete $phl_threads{$name};
              &phl_timer("THREAD_END", "$name")  if ($phl{timer} >= 2);
           } elsif ($phl_fork{$name}) {  # Найден дочерний процесс

              # Считываем из программного канала
              my $fh = $phl_fork{$name};  # Считываем дескриптор канала чтения данных в зависимости от имени дочернего процеса
              @result = <$fh>;
              close $fh;

              # Переменной ставим флаг юникода, если данные utf-8
              if ($phl{encoding_out} eq "utf8") {
                 utf8::decode($_)  foreach (@result);
              }

              $result{$name} = "@result";

              delete $phl_fork{$name};
              &phl_timer("FORK_END", "$name")  if ($phl{timer} >= 2);
           } else {
              &phl_error("Процесс <B>'$name'</B> не найден<br>");
           }
        }

        # если указаны имена одной или нескольких нитей или дочернего процесса возвращаем результаты в виде скаляра или массива - в зависимости от контекста
        if ($wantarray eq 'code') {  #

           return @result;

        } elsif ($wantarray == 0  or  $wantarray == 2) {  # Скалярный и неопределенный контекст
           return ($#result) ? \@result : $result[0];
        } else {

           return %result;
        }

     } else {  # Возвращаем результаты всех нитей и дочерних процессов в виде хеша
        foreach my $name (keys %phl_threads) {  # Обход всех открытых нитей
           my @result = $phl_threads{$name}->join();  # join() => use threads

           $result{$name} = ($#result) ? \@result : $result[0];

           &phl_timer("THREAD_END", "$name")  if ($phl{timer} >= 2);
        }
        %phl_threads = ();

        foreach my $name (keys %phl_fork) {  # Обход всех открытых дочерних процессов
           my $fh = $phl_fork{$name};  # Считываем дескриптор канала чтения данных в зависимости от имени дочернего процеса
           my @result = <$fh>;
           close $fh;

           chomp(@result);

              # Переменной ставим флаг юникода, если данные utf-8
              if ($phl{encoding_out} eq "utf8") {
                 utf8::decode($_)  foreach (@result);
              }

           $result{$name} = ($#result) ? \@result : $result[0];

           &phl_timer("FORK_END", "$name")  if ($phl{timer} >= 2);
        }
        %phl_fork = ();

        return %result;  # Если имена нитей или дочерних процессов не указаны - возвращаем результаты в виде хеша
     }
}

# Функция include()
sub include {

  my $file_include = $_[0];  # включаемый в тело документа файл
  my $param        = $_[1];  # параметры
  my $name_process = $_[2];

  # Считываем кодировку родительского файла
  my $encoding_this_file = $phl_files_decode[-1];

  $file_include = &phl_decode("$encoding_this_file", "$file_include");
  
  &phl_timer("INCLUDE_BEGINE", "$file_include")  if ($phl{timer} >= 2);

  # Запоминаем кодировку "дочернего", загружаемого файла include()
  if ($param =~ /\b(?:dc|decode)\b=*([\w\-]*)/) {
     push @phl_files_decode, "$1";
  } else {
     push @phl_files_decode, "$phl{encoding_in}";  
  }
  
  # Запоминаем "родительский" $ENV{REQUEST_METHOD}
  my $request_method_old = $ENV{REQUEST_METHOD};
  $ENV{REQUEST_METHOD} = "GET";

  my $wantarray;
  if (wantarray()) {  # Списковый контекст
     $wantarray = 1;
  } elsif (defined wantarray()) {  # Скалярный контекст
     $wantarray = 2;
  } else {  # Неопределенный контекст
     $wantarray = 0;
  }

  $wantarray = $_[3]  if ($_[3]);  ## ?

  if ($file_include eq "JOIN") {
     return &phl_end_process($param, $wantarray);  # param ??
  } elsif ($phl_export{include}{$file_include}) {  # Поиск расширения include от сторонних модулей
     # передаем управление подпрограмме, указанной в %phl_export
     $phl_export{include}{$file_include} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
     return;
  }

  if ($param =~ /\b(?:fk|fork|td|thread)\b/) {

     my @result;
     if ($param =~ s/\b(?:fk|fork)\b//g) {
        @result = &phl_fork($file_include, $param, $name_process);
     } elsif ($param =~ s/\b(?:td|thread)\b//g) {
        @result = &phl_thread($file_include, $param, $name_process);
     }

     if ($wantarray == 0  or  $wantarray == 2) {  # Скалярный и Неопределенный контекст
        # Если в скалярном контексте нужно вернуть из доч.процесса список - передается ссылка на массив
        return ($#result) ? \@result : $result[0];
     } else {  # списковый контекст
        return @result;
     }
  }

  my ($dir_include, $file_name, $separator, $suffice) = &phl_fileparse("$file_include");

  my $file_include_cor = $dir_include.$file_name.$suffice;
  
  my ($url_dir, $file_url, $file_path, $file_dir, $http);
  if ($file_include_cor =~ /^(?:http|ftp)[s]{0,1}:/) { # Полный путь

     our %phl_escapes;
     for (0..255) {
        $phl_escapes{chr($_)} = sprintf("%%%02X", $_);
     }
     
     $suffice =~ s/([^A-Za-z0-9\-\._~=&])/$phl_escapes{$1}/ge;

     $file_url = $file_include_cor;
     $http = 'yes';
  } elsif ($file_include_cor =~ m|^/|) { # Относительно домена путь

     $file_url = "http://$ENV{SERVER_NAME}".$file_include_cor;
     $file_dir = $ENV{DOCUMENT_ROOT}.$dir_include;
     $file_path = $file_dir.$file_name;

     $url_dir = "http://$ENV{SERVER_NAME}".$dir_include;

     # Если не находим файл "cgi", пробуем искать в другом месте
     unless (-d "$file_dir") {
        my $script_dir;
        if ($phl{script_dir}) {
           $script_dir = $phl{script_dir};
        } else {
           ($script_dir = $ENV{SCRIPT_FILENAME}) =~ s/$ENV{SCRIPT_NAME}//;
        }

       $file_dir = $script_dir.$dir_include;
       $file_path = $file_dir.$file_name;
     }
  } else { # Относительно файла "*.$phl{pl}" путь

     $file_url = (&phl_fileparse("http://$ENV{SERVER_NAME}$ENV{PATH_INFO}"))[0].$file_include_cor;

     # Узнаем, где мы находимся
     my $phl_path = $phl_path_process[-1];

     $file_dir = $phl_path.$dir_include;
     $file_path = $file_dir.$file_name;

     $url_dir = $phl_path_url[-1].$dir_include;

  }

  if ($http ne "yes") {
 
     if (-d "$file_dir") {
        if (!$file_name) {
           if (-e $phl{index}) {
              # Если находим файл $phl{index} - загружаем
              $file_name = $phl{index};
           } else {
              # Ищем любой файл "index" средствами настроек сервера
              $param .= " socket ";
              $http = "yes";
           }
        } else {
           # Проверяем наличие запрашиваемого файла по указанному пути
           unless (-e "$file_path") {
           }
        }
        $file_path = $file_dir.$file_name;
     } else {
        print "<font color=green>\$file_dir = '$file_dir' НЕ НАЙДЕН</font><BR>";
        &phl_error("Каталог '$file_dir' не найден<br>");  # Выводим ошибку на экран
        return;
     }
  }

  # Поиск параметров от сторонних модулей
  my $finde_param;
  foreach my $name_param (split(' ', $param)) {
     if ($phl_export{param}{$name_param}) {
        $finde_param = $name_param;
        last;
     }
  }

  my @result;
  if (!$file_include) {
     @result = ('test');
  } elsif (!$http  and  $param !~ /\b(?:sk|socket)\b/i) {

     # Кодировка $ENV{'QUERY_STRING'} родительского файла
     if ($encoding_this_file ne "utf8") {
        $suffice = encode("$phl{encoding_out}", "$suffice");
     }

     # Изменяем переменную окружения, добавляя $suffice из $file_include
     if ($suffice) {
        ($ENV{'QUERY_STRING'} = $suffice) =~ s/^\?//;  ### ? Странно, зачем? Но иначе не работает
        $phl_var{'QUERY_STRING'} = $ENV{'QUERY_STRING'};
     }
     
     # Запоминаем переменную окружения в массиве
     push @phl_query_string, $phl_var{'QUERY_STRING'};
     
     # Запоминаем хэш по данным POST и GET
     my %phl_query_copy = %phl_query;
     &phl_query("include");

     unshift(@INC, "$file_dir"); # Добавляем путь к списку каталогов поиска модулей

     chdir "$file_dir" or die "Невозможно cd в '$file_dir': $!\n";  # Изменяем текущий каталог процесса

     # Запоминаем данные по текущему каталогу в массив
     push @phl_path_process, $file_dir;

     push @phl_path_url, $url_dir;

     if ($finde_param  and  $phl_export{param}{$finde_param}) { # если параметр совпадает с ключем хэша %phl_export

        # передаем управление подпрограмме, указанной в %phl_export
        if ($wantarray == 1) {  # Списковый контекст
           @result = $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
        } elsif ($wantarray == 2) {  # Скалярный контекст
           my $result = $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
           @result = ($result);
        } else {  # Неопределенный контекст
           $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
        }

     } elsif ($param =~ /\b(?:on|open)\b/
              or
              ($file_name =~ /\.htm[l]+$/  and  $param !~ /\b(?:phl)\b/)) {

        my $decode;
        if ($param =~ /\b(?:dc|decode)\b=*([\w\-]*)/) {
           $decode = $1;
        } else {
           $decode = $phl{encoding_in};
        }

        if ($wantarray == 1) {  #списковый контекст
           open(my $fh, "$file_name")  or  &phl_error("$dir_include$file_name", '$!');
              flock($fh, 1);
              @result = <$fh>;
           close($fh);
         
           foreach (@result) {              
              if ($decode eq "utf8") {
                 utf8::decode($_);
              } elsif (!$decode) {
              } else {
                 $_ = decode($decode, $_);
              }
           }
           
        } else {  # Скалярный и неопределённый контекст
           my $code;
           open(my $fh, "<$file_name")  or  &phl_error("$dir_include$file_name", '$!');
              while (sysread($fh, my $bate, 8*1024)) {
                 $code .= $bate;
              }
           close($fh);
         
           if ($decode eq "utf8") {
              utf8::decode($code);
           } elsif (!$decode) {
           } else {
              $code = decode($decode, $code);
           }
         
           $code = &phl_abs($code, "$file_url")  if ($param  and  $param =~ /\babs\b/);

           @result = ($code);
        }

        if ($phl{encoding_out} ne "utf8") {
           foreach (@result) {
              $_ = encode($phl{encoding_out}, $_);
           }
        }
        
        if ($wantarray == 0) {  # Неопределенный контекст
           print @result;  # Выводим результат на экран
           $wantarray = -1;
        }

     } elsif ($file_name =~ /\.cgi$/  or  $param =~ /\bcgi\b/) {

        my $cgi_code;
        if ($param =~ /\b(?:rq|require)\b/) {
           require "$file_path";  # Проблемы с кодировками
           @result = (1);
        } else {
           open(my $fh, "$file_path") or &phl_error("$file_path", '$!');
              my $bate;
              while (sysread($fh, $bate, 8*1024)) {
                 $cgi_code .= $bate;
              }
           close($fh);

           if ($param =~ /\b(?:dc|decode)\b=*([\w\-]*)/) {
              my $decode = $1;
              $cgi_code = decode($decode, $cgi_code);
           } elsif ($phl{encoding_in}) {
              $cgi_code = decode($phl{encoding_in}, $cgi_code);
           }

           @result = &phl_code_perl_eval($cgi_code, $param, "", "", "include");
        }

     } else {

        if ($param =~ /\b(?:ex|exist)\b/) {
           my $return = (-e "$file_name") ? 1 : 0;  # Возврашаем "1" если файл существует и "0" - если нет
           @result = ($return);
        } elsif ($param =~ /\b(?:hd|head)\b/) {
           my @array = &phl_head("$file_name");
           @result = @array;  # Возврашаем список из 5 элементов => (Content type, Document length, Modified time, Expires, Server)
        } else {
           my $phl_path_include = (&phl_fileparse("$file_path"))[0];

           my $analys = &phl_analys_file($file_path, "$param", "$file_url");
           $analys = &phl_abs($analys, "$file_url")  if ($param  and  $param =~ /\babs\b/);

           @result = &phl_code_perl_eval($analys, $param, $name_process, $file_include, "include");
        }
     }

     # Удаляем последний элемент массива данных по текущему каталогу процессов
     pop @phl_path_process;

     pop @phl_path_url;

     my $phl_path = $phl_path_process[-1];
     chdir "$phl_path" or die "Невозможно cd $phl_path";  # Возвращаемся в текущий каталог процесса

     # Удаляем последний элемент массива, возвращая текущее значение $ENV{'QUERY_STRING'}
     pop @phl_query_string;
     $phl_var{'QUERY_STRING'} = $phl_query_string[-1];

     # Возвращаем данные обратно
     %phl_query = %phl_query_copy;

  } else {

     # Кодируем URL, для использования русского написания файлов   

     if ($finde_param  and  $phl_export{param}{$finde_param}) { # если параметр совпадает с ключем хэша %phl_export
        # передаем управление подпрограмме, указанной в %phl_export
        if ($wantarray == 1) {  # Списковый контекст
           @result = $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
        } elsif ($wantarray == 2) {  # Скалярный контекст
           my $result = $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
           @result = ($result);
        } else {  # Неопределенный контекст
           $phl_export{param}{$finde_param} -> ("$file_include", "$param", "$name_process", "$file_url", "$file_name");
        }
     } elsif ($param =~ /(?:^| )(?:exist|ex)(?: |$)/) {
        @result = &phl_get_socket("$file_url", "exist");

     } elsif ($param =~ /(?:^| )(?:head|hd)(?: |$)/) {
        # Возврашаем массив заголовков
        @result = &phl_get_socket("$file_url", "head");  # 
     } else {
        # Возвращаем тело запрашиваемого документа с учетом параметров, указанных в $param
        @result = &phl_get_socket("$file_url", "", "$param");   # &phl_get_socket использует IO::Socket::INET->new() <= use IO::Socket::INET

        if ($wantarray == 0) {  # Неопределенный контекст
           print @result;  # Выводим результат на экран
           $wantarray = -1;
        }
     }
        
  }
  
  # Возвращаем первоначальный $ENV{REQUEST_METHOD}, вызвавшему include
  $ENV{REQUEST_METHOD} = $request_method_old;

  &phl_timer("INCLUDE_END", "$file_include")  if ($phl{timer} >= 2);

  # Удаляем последний элемент массива, возвращая текущее значение кодировки
  pop @phl_files_decode;

  if ($wantarray == 0  or  $wantarray == 2) {  # Скалярный и Неопределенный контекст
      return undef unless (@result);
      return $result[0];
  } elsif ($wantarray == 1) {  # списковый контекст
      return @result;
  } else {
      return;  # Не нужно передавать никаких данных, к примеру, если печатаем на экран
  }

}

sub refs {
  no strict "refs";
  my $data = $_[0];
  $data =~ s/^\$(.+)/$1/;
  return ${$data};
}

# Подпрограмма по получению документа (используется модуль 'IO::Socket::INET')
sub phl_get_socket {

  my $file_url = $_[0];   
  my $head     = $_[1];
  my $param    = $_[2];
  
  my ($http, $host, $port, $path) = $file_url =~ m{^((?:http|ftp)[s]{0,1})://([^/:\@]+)(?::(\d+))?(/\S*)?\s*$};

  &phl_timer("use IO::Socket::INET/SSL", "$file_url")  if ($phl{timer} >= 4);

  my $sock;
  if ($http eq "https") {
     eval "use IO::Socket::SSL";
     &phl_timer("begin use IO::Socket::INET/SSL", "$file_url")  if ($phl{timer} >= 4);
     $sock = IO::Socket::SSL->new(
        PeerHost => "$host",
        PeerPort => "https",
     ) or return undef;
  } else {
     $port ||= 80;
     eval "use IO::Socket::INET";
     &phl_timer("end use IO::Socket::INET/SSL", "$file_url")  if ($phl{timer} >= 4);
     $sock = IO::Socket::INET->new(
        PeerAddr => "$host",  
        PeerPort => $port,
        Proto    => 'tcp',
        Timeout  => 60
     ) or return undef;
     
     $host .= ":$port"  if $port != 80;
  }

  &phl_timer("new IO::Socket::INET/SSL", "$_[0]")  if ($phl{timer} >= 4);

  $sock -> autoflush;
  $path = "/" unless defined($path);

print $sock <<"[END]";
GET $path HTTP/1.0
Host: $host
User-Agent: Perl_use_socket_for_phl/0.01
User: Phl.pm

[END]

  my ($content, $header, $http_200);
  my ($line, $error_300, $location, $charset);
  while (defined ($line = <$sock>)) {
     
     # Анализ заголовка
     if ($line =~ /^HTTP/) {
        if ($line =~ /200/) {
           $http_200 = "ok";
        } elsif ($line =~ /30/) {
           $error_300 = "yes";
        }
     # Переадресация при ошибках 300
     } elsif ($line =~ m/Location: (\S+)/i) {
        $location = $1;
     # Определяем кодировку
     } elsif ($line =~ /charset=([\w\-]*)/i) {
        $charset = $1;
     }
     
     $header .= $line;
     last  if ( $line =~ /^\r/  or $line =~ /^\r?\n/);
  }
  
  # Обработка ошибок
  if ($error_300 and $location) {
     # Если переадресация ведет на другой адрес, не содержащий $path
     if ($location !~ /\Q$path\E$/) {
        if ($head  and  $head eq "exist") {
           close ($sock);
           return (0);        
        } elsif ($head  and  $head eq "head") {
           close ($sock);
           my @header = split(/\n/, $header);
           return @header;
        }
     }
  }
  
  if ($error_300 and !$_[3]) {
     close ($sock);
     my @content = &phl_get_socket("$location", "$head", "$param", "location");  
     return @content;
  } elsif ($error_300 and $_[3]) {
     close ($sock);
     my @content = &phl_get_socket("$location", "$head", "$param");  
     return @content;
  } elsif ($head  and  $head eq "head") {
     close ($sock);
     my @header = split(/\r?\n/, $header);
     return @header;
  } elsif ($head  and  $head eq "exist") {
     close ($sock);
     if ($http_200) {
        return (1);
     } else {
        return (0);
     }
  }
  
  # Тело документа
  while (read($sock, my $buffer, 512)) {
     $content .= $buffer;
  }
     
  close ($sock);

  &phl_timer("GET_FILE_FROM_Socket", "$_[0]")  if ($phl{timer} >= 4);

  # Декодируем загружаемый файл в зависимости от кодировки, указанной в $param и настройках
  if ($param =~ /\b(?:dc|decode)\b=*([\w\-]*)/) {
     my $decode = $1;
     if ($decode) {
        $content = &phl_decode("$decode", "$content");
     } elsif ($charset) {
        $content = &phl_decode("$charset", "$content");
     } else {
        $content = &phl_decode("$phl{encoding_in}", "$content");
     }
     # Переводим кодировку из внутреннего формата в заданный в $phl{encoding_out}
     $content = &phl_encode("", "$content");

  } elsif ($param =~ /\b(?:nd|no_decode)\b/) {
     # Ничего никуда не декодируем
  } else {
     if ($charset) { # Если кодировка указана в заголовке HTTP
        $content = &phl_decode("$charset", "$content");
     } else {
        # Ищем указание на кодировку в META
        my ($charset_http) = $content =~ /<meta .*?charset=["']*([\w-]+)["']*\s*.*?>/i;
        $content = &phl_decode("$charset_http", "$content");
     }

     # Переводим кодировку из внутреннего формата в заданный в $phl{encoding_out}
     $content = &phl_encode("", "$content");
  }
		
  $content = &phl_abs($content, "$file_url")  if ($param !~ /(?:na|no_abs)/);
   
  return ($content);
}

sub phl_abs {

   my $content = $_[0];
   my $dir = $_[1];

   $content =~ s/<(a |img |script |form |link |frame |area )(.*?)(href|src|action)=("|')(.*?)(\4)/"<$1$2$3=$4".phl_new_abs($5, $dir, $4, $6)."$6"/gie;
   $content =~ s/<(a |img |script |form |link |frame |area )(.*?)(href|src|action)=()([^"'].*?)( |>|\n)/"<$1$2$3=$4".phl_new_abs($5, $dir, $4, $6)."$6"/gie;

   return $content;
}

sub phl_new_abs{
  my $link = shift;
  my $address = shift;
  my $l = shift;
  my $r = shift;
  
  my ($dir, $file_name, $separator, $suffice) = &phl_fileparse("$address");
  
  my $url;
  if ($link =~ /^(?:http|https|ftp):/i) { # Полный путь
     $url = $link;
  } elsif ($link =~ m|^//|) { # Относительно домена путь
     $url = "http:".$link;
  } elsif ($link =~ m|^/|) { # Относительно домена путь
     (my $dom = $dir) =~ s{(http|https|ftp)://(.*?)/.*}{$1://$2};
     $url = $dom.$link;
  } else { # Относительно файла "*.$phl{pl}" путь
     $url = $dir.$link;
  }
 
  return "$url";
}

sub phl_head {
   my $file_name = $_[0];
   my @stat_file = stat($file_name);
   return @stat_file;
}

# Кодируем строку запроса для передачи по http
sub phl_escape {
   my $line = $_[0];
   our %escapes;
   $escapes{chr($_)} = sprintf("%%%02X", $_)  for (0..255);
   $line =~ s/([^A-Za-z0-9\-\._~\/=&:?])/$escapes{$1}/ge;  # Переделать, разбив запрос на домен, адрес файла и остальное
   return $line;
}

BEGIN {
  # Метка времени конца компиляции программы
  $phl_var{phl_t1} = gettimeofday;  # gettimeofday <= use Time::HiRes
  &phl_timer("Конец компиляции", "Phl.pm")  if ($phl{timer} >= 2);
}

1;

